:- use_module( library(lists) ).
:- use_module( library(ugraphs) ).
:- use_module( library(ordsets) ).

parents(VERTEX, GRAPH, NEIGHBORS) :-
	neighbours(VERTEX, GRAPH, NEIGHBORS).

children(VERTEX, GRAPH, CHILDREN) :-
	findall(V,(parents(V,GRAPH,As),ord_member(VERTEX,As)),CHILDREN).
% Definition of what a Bayes net is...


/*---------------------------------------------------------------------
  Predicate: bn(+RVs,-BN,-AncBN)
       Args: RVs = a list of random variables
             BN =  a DAG with the RVs as vertices
	     AncBN = a DAG where a node's ancestors in BN are it's parents.
   Succeeds: 2^O(n^2*log n) where n is length of RVs
   Comments: The DAG represents a BN
             The BN need not be fully connected
             AncBN is used for quick checking for cycles
---------------------------------------------------------------------*/

bn([],[],[]).
bn([RV|RVs],BN,AncBN) :-                              %it's a Bayes net if
	bn(RVs,TailBN,AncTailBN),                     %this is and ..
	connect(RV,RVs,TailBN,AncTailBN,BN,AncBN).    %we connect RV thus giving BN ...


/*---------------------------------------------------------------------
  Predicate: connect(+RV,+RVs,+TailBN,+AncTailBN,-BN,-AncBN)
       Args: RV = a random variable
             RVs = a list of random variables
             TailBN =  a DAG with the RVs as vertices
             AncTailBN =  ancestor version of TailBN
             BN = a directed graph produced by adding RV to TailBN
             AncBN = ancestor version of BN
   Succeeds: 3^n, where n is length of RVs
   Comments: BN may contain cycles             
---------------------------------------------------------------------*/

connect(RV,RVs,TailBN,AncTailBN,BN,AncBN) :-
	choose_edges(RVs,RV,Edges),              %pbc choice
	\+ cycle_introduced(AncTailBN,RV,Edges), %check that no cycle has been created
	add_vertices(TailBN,[RV],BN1),           %RV gets added, even if Edges=[]
	add_vertices(AncTailBN,[RV],AncBN1),     %RV gets added, even if Edges=[]
	add_edges(BN1,Edges,BN),                 %from library
	add_anc_edges(Edges,RV,AncBN1,AncBN).       %update ancestral sets

% % find all vertices which have child as an ancestor and add RV
% % as yet another one.
add_anc_edges([Child-RV|Edges],RV,AncBN1,AncBN) :-
	Child \= RV,
	children(Child,AncBN1,Descendants),             % grab the child's descendants
	construct_child_edges(Descendants,RV,DescEdges),% construct an edge to RV from each of these
	add_edges(AncBN1,[Child-RV|DescEdges],AncBN2),  % add them to ancestral DAG
	add_anc_edges(Edges,RV,AncBN2,AncBN).           % and continue

add_anc_edges([],_RV,AncBN,AncBN).
add_anc_edges([RV-Parent|Edges],RV,AncBN1,AncBN) :-
	% !,                                            % since never get RV-RV
	parents(Parent,AncBN1,Ancestors),             % grab the parents ancestors
	construct_parent_edges(Ancestors,RV,AncEdges),% construct an edge from RV to each of these
	add_edges(AncBN1,[RV-Parent|AncEdges],AncBN2),% add them to ancestral DAG
	add_anc_edges(Edges,RV,AncBN2,AncBN).         % and continue

% each element of list is a parent of RV
construct_parent_edges([],_RV,[]).
construct_parent_edges([H|T],RV,[RV-H|T1]) :-
	construct_parent_edges(T,RV,T1).
	
% each element of list is a child of RV
construct_child_edges([],_RV,[]).
construct_child_edges([H|T],RV,[H-RV|T1]) :-
	construct_child_edges(T,RV,T1).
	

cycle_introduced(AncTailBN,RV,Edges) :-                % RV has introduced a cycle if ...
	member(RV-Parent,Edges),                       % Parent is a parent of RV 
	parents(Parent,AncTailBN,ParentsAncestors),    % Ancestors are the ancestors of Parent
	member(Child-RV,Edges),                        % Child is a child of RV
	ord_member(Child,ParentsAncestors).            % and Child is an ancestor of Parent!
	
	



/*---------------------------------------------------------------------
  Predicate: choose_edges(+RVs,+RV,-Edges)
       Args: RVs = a list of random variables
             RV = a random variable
	     Edges = a list of edges connecting RV to members of RVs 
   Succeeds: 3^n, where n is length of RVs
   Comments: RVs may be empty
---------------------------------------------------------------------*/

choose_edges([],_RV,[]).
choose_edges([RVsH|RVsT],RV,Edges) :-
	choose_edges1([RVsH|RVsT],RV,Edges).

/*---------------------------------------------------------------------
  Predicate: choose_edges1(+RVs,+RV,-Edges)
       Args: RVs = a list of random variables
             RV = a random variable
	     Edges = a list of edges connecting RV to members of RVs 
   Succeeds: 3^n, where n>0 is length of RVs
   Comments: Fails if n=0 
---------------------------------------------------------------------*/

0.3 :: choose_edges1([H|T],RV,[H-RV|Rest]) :-  %RV parent of H
	choose_edges(T,RV,Rest).
0.4 :: choose_edges1([H|T],RV,[RV-H|Rest]) :-  %RV child of H
	choose_edges(T,RV,Rest).
0.3 :: choose_edges1([_H|T],RV,Rest) :-         %no direct connection
	choose_edges(T,RV,Rest).

/*---------------------------------------------------------------------
  Predicate: no_cycle_introduced(+_RV,+BN)
       Args: RV = a random variable
	     BN = a directed graph
   Succeeds: if BN is acyclic
   Comments: stupidly does not take advantage of the fact that
             BN without RV is guaranteed to be acyclic.
---------------------------------------------------------------------*/

/* 
no_cycle_introduced(_RV,BN) :-         %no cycles if there is a top sort
	top_sort(BN,_).                %from library
*/
