:- use_module( library(lists) ).
:- use_module( library(ugraphs) ).
:- use_module( library(ordsets) ).



/*---------------------------------------------------------------------
  Predicate: graphs(+RVs,-Skeleton,-EssentialGraph,-Imms,-BN)
       Args: 
             
   Succeeds: 
   Comments: graphs(RVs,Skeleton,EssentialGraph,Imms,BN) is true if:
              Skeleton is an undirected graph with RVs for vertices and
	      EssentialGraph is a chain graph with Skeleton as a skeleton and
	      Imms as its immoralities and
	      BN is a Bayes net formed by directing undirected edges in EssentialGraph
	      such that it has the same immoralities as EssentialGraph 
---------------------------------------------------------------------*/

bn(RVs,BN) :-
	graphs( RVs, _S, _E, _I, BN ).

graphs(RVs,Skeleton,EssentialGraph,Imms,BN) :-
	skeleton(RVs,Skeleton),                %choose undirected skeleton
	essential_graph(Skeleton,Imms,EssentialGraph), %choose "essential graph"
	bn(EssentialGraph,Imms,BN),             %choose directed graph
	top_sort(BN,_).                        %check it's acyclic


bn(ChainGraph,Imms,BN) :-
	edges(ChainGraph,Edges),
	edges_to_delete(Edges,Togo),
	del_edges(ChainGraph,Togo,BN),
	immoralities(BN,Imms).


edges_to_delete([],[]).
edges_to_delete([V1-V2|InT],[ChosenOne|OutT]) :-
	ord_member(V2-V1,InT),             %one has to go
	!,
	ord_del_element(InT,V2-V1,NewT),
	binary_choice(V1-V2,V2-V1,ChosenOne),
	edges_to_delete(NewT,OutT).
edges_to_delete([_|TIn],TOut) :-	
	  edges_to_delete(TIn,TOut).

binary_choice(X,_Y,X).  %add labels for SLP
binary_choice(_X,Y,Y).

	  
/*---------------------------------------------------------------------
  Predicate: skeleton(+RVs,-Skeleton)
       Args: RVs = a list of random variables
             Skelton =  an (undirected) graph with the RVs as vertices
   Succeeds: n*(n-1)/2 where n is length of RVs
   Comments: This is just one way of choosing a skeleton
---------------------------------------------------------------------*/

skeleton([],[]).
skeleton([RV|RVs],Skeleton) :-                             
	skeleton(RVs,TailSkeleton),
	connect(RV,RVs,TailSkeleton,Skeleton).    


/*---------------------------------------------------------------------
  Predicate: immoralities(+Skeleton,-Immoralities,-ChainGraph)
       Args: 
             
   Succeeds: 
   Comments: 
---------------------------------------------------------------------*/

essential_graph(Skeleton,Immoralities,ChainGraph) :-
	all_potential_immoralities(Skeleton,Triples),  %find all potential immoralities
	choose_from_list(Triples,Immoralities),             %choose some
	impose_immoralities(Immoralities,Skeleton,ChainGraph).


impose_immoralities([],ChainGraph,ChainGraph).
impose_immoralities([Mum-Child-Dad|T],In,Out) :-
	del_edges(In,[Mum-Child,Dad-Child],Mid),
	impose_immoralities(T,Mid,Out).



all_potential_immoralities(Skeleton,Triples) :-
	vertices(Skeleton,Vertices),
	findall(Mum-Child-Dad,
	(
	    member(Child,Vertices),
	    neighbours(Child,Skeleton,Neighbours),
	    append(_,[Mum|T],Neighbours),            %Mum always before Dad in
	    neighbours(Mum,Skeleton,MumsNeighbours), %standard order
	    member(Dad,T),
	    \+ ord_member(Dad,MumsNeighbours)
	),
	Triples).

immoralities(Skeleton,Triples) :-
	vertices(Skeleton,Vertices),
	findall(Mum-Child-Dad,
		(
		  member(Child,Vertices),
		  neighbours(Child,Skeleton,Neighbours),
		  append(_,[Mum|T],Neighbours),	%Mum always before Dad in
		  neighbours(Mum,Skeleton,MumsNeighbours), %standard order
		  member(Dad,T),
		  \+ ord_member(Dad,MumsNeighbours),
		  neighbours(Dad,Skeleton,DadsNeighbours),
		  \+ ord_member(Mum,DadsNeighbours)
		  ),
		Triples).

choose_from_list([],[]).
choose_from_list([H|T],Choice) :-
	choose_from_list1([H|T],Choice).


%add labels to make SLP	
choose_from_list1([_H|T],Choice) :-    %don't include H
	choose_from_list(T,Choice).
choose_from_list1([H|T],[H|Choice]) :- %do    include H
	choose_from_list(T,Choice).

	


/*---------------------------------------------------------------------
  Predicate: connect(+RV,+RVs,+TailSkeleton,-Skeleton)
       Args: 
             
   Succeeds: 
   Comments: 
---------------------------------------------------------------------*/

connect(RV,RVs,TailSkeleton,Skeleton) :-
	choose_edges(RVs,RV,Edges),              %pbc choice
	add_vertices(TailSkeleton,[RV],S),       %RV gets added, even if Edges=[]
	add_edges(S,Edges,Skeleton).             %from library



/*---------------------------------------------------------------------
  Predicate: choose_edges(+RVs,+RV,-Edges)
       Args: RVs = a list of random variables
             RV = a random variable
	     Edges = a list of edges connecting RV to members of RVs 
   Succeeds: 2^n, where n is length of RVs
   Comments: RVs may be empty
---------------------------------------------------------------------*/

choose_edges([],_RV,[]).
choose_edges([RVsH|RVsT],RV,Edges) :-
	choose_edges1([RVsH|RVsT],RV,Edges).

/*---------------------------------------------------------------------
  Predicate: choose_edges1(+RVs,+RV,-Edges)
       Args: RVs = a list of random variables
             RV = a random variable
	     Edges = a list of edges connecting RV to members of RVs 
   Succeeds: 2^n, where n>0 is length of RVs
   Comments: Fails if n=0 
---------------------------------------------------------------------*/

%add labels to make into an SLP

choose_edges1([H|T],RV,[H-RV,RV-H|Rest]) :-  %there's an edge
	choose_edges(T,RV,Rest).
choose_edges1([_H|T],RV,Rest) :-             %no edge
	choose_edges(T,RV,Rest).

