% :- module( llhood, 

:- pl( sicstus(_S), ensure_loaded( logamma_sicstus ) ).
:- pl( yap(_Y), reconsult( logamma_yap ) ).

/* 
From jc@cs.york.ac.uk Tue Mar  6 15:53:27 2001
Date: Tue, 6 Mar 2001 11:31:49 GMT
From: jc@cs.york.ac.uk
To: nicos@cs.york.ac.uk

*/

/***********************************************************************

Written Tue Mar  6 11:19:36 GMT 2001

Only llhood_ratio/3 is called from outside
This computes the log of the likelihood ratio of 2 BNs

BNs look like this:

bn(1,[1-[5],2-[],3-[1,5],4-[2],5-[4]]).
bn(2,[1-[5],2-[1,5],3-[4],4-[1],5-[]]).
bn(3,[1-[2,5],2-[5],3-[1,4,5],4-[1,5],5-[]]).

Data (not in this file) looks like this:
data(0,0,1,1,0,1,1,0).

Only works for data in data/8 format

***********************************************************************/

:- use_module(library(ugraphs)).
:- use_module(library(lists)).
:- use_module(library(terms)).
:- use_module(library(ordsets)).

:- dynamic cached_family_contrib/3.


% :- 	user:datafile( Data ),
	% user:write( using_data(Data) ), nl,
	% compile( Data ).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%START COMPUTATION OF LOG OF LIKELIHOOD RATIO%%%%%%%%%%%%%%%%%%%%%%%%%

/*---------------------------------------------------------------------
  Predicate: llhood_ratio(+BN1,+BN2,-LogofRatio)
       Args: BN1 = Bayesian net as a list of familiies
             BN2 = Bayesian net as a list of familiies
             LogofRatio = log(P(data|BN1)/P(data|BN2))
   Succeeds: 1
   Comments: This is a function
---------------------------------------------------------------------*/

llhood_ratio(BN1,BN2,LogofRatio) :-
	llhood_ratio(BN1,BN2,0,LogofRatio).

/*---------------------------------------------------------------------
  Predicate: llhood_ratio(+BN1,+BN2,+Acc,-LogofRatio)
       Args: BN1 = Bayesian net as a list of familiies
             BN2 = Bayesian net as a list of familiies
	     Acc = accumulator
             LogofRatio = Acc + log(P(data|BN1)/P(data|BN2))
   Succeeds: 1
   Comments: This is a function
             if families are identical then both cancel out, and can be ignored
---------------------------------------------------------------------*/

llhood_ratio([],[],L,L).                 
llhood_ratio([Child-Parents|Rest1],[Child-Parents|Rest2],In,Out) :- %identical families
	!,
	llhood_ratio(Rest1,Rest2,In,Out).
llhood_ratio([Family1|Rest1],[Family2|Rest2],In,Out) :-
	llhood_family_contrib(Family1,Family_Contrib1),
	llhood_family_contrib(Family2,Family_Contrib2),
	Mid is In + Family_Contrib1 - Family_Contrib2,
	llhood_ratio(Rest1,Rest2,Mid,Out).

/*---------------------------------------------------------------------
  Predicate: llhood_family_contrib(+Family,-Contrib)
       Args: Family = Child-[Parent1,Parent2,...] 
             Contrib = Family's contribution to the log-likelihood
   Succeeds: 1
   Comments: Computes the 'score' for each family
             ie its contribution to the log-likelihood
             Either grab cached value or compute it
             This is a function
---------------------------------------------------------------------*/

llhood_family_contrib(Family,Contrib) :-
	term_hash(Family,HashValue),                
	(
	    cached_family_contrib(HashValue,Family,Contrib) ->
	    
	    true ;
	    
	    counts_for_family(Family,ChildCountsList),
	    compute_llhood_family_contrib(ChildCountsList,Family,0,Contrib),
	    assert(cached_family_contrib(HashValue,Family,Contrib))
	).

/*---------------------------------------------------------------------
  Predicate: cached_family_contrib(+Family,-Contrib)
       Args: Family = Child-[Parent1,Parent2,...] 
             Contrib = Family's contribution to the log-likelihood
   Succeeds: 1
   Comments: Caches scores for future retreival
             No definition, it gets asserted!
---------------------------------------------------------------------*/


%%%%%%%%%%%%%%%%%%%%%%%%%%%%END COMPUTATION OF LOG OF LIKELIHOOD RATIO
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%




%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%START PREDICATES FOR GETTING COUNTS%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

/*---------------------------------------------------------------------
  Predicate: counts_for_family(+Family,-ChildCountsList)
       Args: Family = Child-[Parent1,Parent2,...] 
             ChildCountsList = for each configuration of the parents
	     a list of labelled counts for child 
   Succeeds: 1
   Comments: This is a function
             Each element of ChildCountsList is something like [1-20,2-30,3-50]-[1,3]
	     Child counts are for Parent1=1, Parent2=3
             [1-20,2-30,3-50] means that the child took value=1 20 times, etc
---------------------------------------------------------------------*/

counts_for_family(Child-Parents,ChildCountsList) :-
	% Goal = data(_,_,_,_,_,_,_,_),
	data_format( Goal ),
	bind([Child|Parents],Goal,[ChildVar|PVars]),
	findall(PVars-ChildVar,Goal,Instances),      % constructs massive list
	keysort(Instances,SortedInstances),          %order by values of parents
	child_counts_list(SortedInstances,[],ChildCountsList).


/*---------------------------------------------------------------------
  Predicate: bind(+ArgPositions,?Term,+TermList)
       Args: ArgPositions = list of argument positions
             Term = any term
	     TermList = list of subterms of Term
   Succeeds: 1 (unless ArgPositions has an arg outside arity of Term, then fail)
   Comments: This is a function
             Yanks out relevant arguments from Goal and puts them in TermList
---------------------------------------------------------------------*/

bind([],_,[]).
bind([H|T],Goal,[H1|T1]) :-
	arg(H,Goal,H1),
	bind(T,Goal,T1).

/*---------------------------------------------------------------------
  Predicate: child_counts_list(+SortedInstances,+SoFar,-ChildCountsList)
       Args: SortedInstances = sorted instances of a family
             SoFar = accumulator for counts
	     ChildCountsList = list of counts for child, one per parent configuration
   Succeeds: 1
   Comments: This is a function
             Each element of ChildCountsList is something like [1-20,2-30,3-50]-[1,3]
	     Child counts are for Parent1=1, Parent2=3
             [1-20,2-30,3-50] means that the child took value=1 20 times, etc
---------------------------------------------------------------------*/

child_counts_list([],ChildCountsList,ChildCountsList).
child_counts_list([PVals-ChildVal|T],SoFar,ChildCountsList) :-
	child_counts([PVals-ChildVal|T],PVals,[],ChildCounts,Rest),
	child_counts_list(Rest,[ChildCounts-PVals|SoFar],ChildCountsList).


/*---------------------------------------------------------------------
  Predicate: child_counts(+SortedInstances,+PVals,+SoFar,-ChildCounts,-Rest)
       Args: SortedInstances = sorted instances of a family
             PVals = parent configuration
             SoFar = accumulator for counts
	     ChildCounts = list of counts for child, for the PVals parent configuration
	     Rest = Counts for other parent configurations
   Succeeds: 1
   Comments: This is a function
---------------------------------------------------------------------*/

child_counts([PVals-ChildVal|T],PVals,Sofar,Out,Rest) :-        %got a PVals match
	!,
	(
	    select(ChildVal-Count,Sofar,Residue) ->             %if seen this ChildVal before
	    C is Count + 1,                                     %increment count
	    NewSofar = [ChildVal-C|Residue] ;                   %put back
	    NewSofar = [ChildVal-1|Sofar]                       %else start the count for this ChildVal
	),
	child_counts(T,PVals,NewSofar,Out,Rest).
child_counts(L,_PVals,Out,Out,L).                               %match failed, we're done with this PVals

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%END PREDICATES FOR GETTING COUNTS
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%START COMPUTATION OF FAMILY LOG-LIKELIHOOD%%%%%%%%%%%%%%%%%%%%%%%%%%%


/*---------------------------------------------------------------------
  Predicate: compute_llhood_family_contrib(+ChildCountsList,+Family,+In,-Out)
       Args: ChildCountsList = list of counts for child, one per parent configuration
             Family = ChildName-List of parent names
             In = accumulator input, initially 0
             Out = 'score' for the family
   Succeeds: 1
   Comments: This is a function
             Each element of ChildCountsList is something like [1-20,2-30,3-50]-[1,3]
	     Child counts are for Parent1=1, Parent2=3
             [1-20,2-30,3-50] means that the child took value=1 20 times, etc
             middle loop of Heckerman's (32)
---------------------------------------------------------------------*/

compute_llhood_family_contrib([],_Family,Out,Out).
compute_llhood_family_contrib([ChildVals-ParentVals|T],Family,In,Out) :-
	config_contrib(ChildVals,Family,ParentVals,In,0,0,Mid),
	compute_llhood_family_contrib(T,Family,Mid,Out).


/*---------------------------------------------------------------------
  Predicate: config_contrib(+ChildVals,+Family,+ParentVals,+In,+AIn,+ANIn,-Out)
       Args: ChildVals = labelled counts for the child
             Family = names of child and parents
	     ParentVals = the values of the parents for this configuration
             In = main accumulator input
             AIn = accumulator for Alpha_ij
             ANIn = accumulator for Alpha_ijplusN_ij
             Out = the result!
   Succeeds: 1
   Comments: This is a function
             inner loop of Heckerman's (32)
            Suppose variable x has parents y and z and when y=1,z=3 the counts for x are
            x=1 20 times, x=2 30 times and x=3 50 times, then goal would be:
            config_contrib([1-20,2-30,3-50],x-[y,z],[1,3]In,0,0,Out)
            we just add two accumulators for the sum of alphas and sum of alphas+counts
---------------------------------------------------------------------*/

config_contrib([],_Family,_ParentVals,In,Alpha_ij,Alpha_ijplusN_ij,Out) :-
	logamma(Alpha_ij,Tmp),
	logamma(Alpha_ijplusN_ij,Tmp2),
	Out is Tmp - Tmp2 + In.
config_contrib([ChildVal-N_ijk|T],Family,ParentVals,In,AIn,ANIn,Out) :-
	alpha(Family,ChildVal,ParentVals,Alpha_ijk),
	Tmp is N_ijk + Alpha_ijk,
	logamma(Tmp,Tmp2),
	logamma(Alpha_ijk,Tmp3),
	Mid is In + Tmp2 - Tmp3,
	AMid is AIn + Alpha_ijk,
	ANMid is ANIn + Tmp,
	config_contrib(T,Family,ParentVals,Mid,AMid,ANMid,Out).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%END COMPUTATION OF FAMILY LOG-LIKELIHOOD


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%START DEFINITION OF PARAMETER PRIORS%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

/*---------------------------------------------------------------------
  Predicate: alpha(+Family,+ChildVal,+ParentVals,-Alpha_ijk)
       Args: Family = names of child and parents
	     ChildVal = a value for the child (j)
             ParentVals = values of the parents (k)
             Alpha_ijk = approprate Dirichlet parameter
   Succeeds: 1
   Comments: This is a function
---------------------------------------------------------------------*/

alpha(_Family,_ChildVal,_ParentVals,1).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% END DEFINITION OF PARAMETER PRIORS
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%





































/**********************************************************************
NOT USED
NOT USED
NOT USED

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% START COMPUTATION OF LIKELIHOOD
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

llhood(BN,L) :-
	llhood(BN,0,L).

% for each family
% does outermost loop of Heckerman's (32)

llhood([],L,L).
llhood([Family|Families],In,Out) :-
	llhood_family_contrib(Family,Family_Contrib),   %this is the 'score' for this family
	Mid is In + Family_Contrib,
	llhood(Families,Mid,Out).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% END COMPUTATION OF LIKELIHOOD
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%



**********************************************************************/
